{
Form WordMaze
-------------

Copyright (c) 2003

Licensed under the BSD/MIT License:

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
}

unit WordMaze;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, ActnList, StrUtils, Printers, ExtCtrls, GenMaze;

type
  TWordMazeForm = class(TForm)
    MainMenu: TMainMenu;
    PopupMenu: TPopupMenu;
    ActionList: TActionList;
    FilePrint: TAction;
    FileExit: TAction;
    FileMenuMain: TMenuItem;
    PrintMain: TMenuItem;
    N1: TMenuItem;
    ExitMain: TMenuItem;
    PrintPopup: TMenuItem;
    ExitPopup: TMenuItem;
    PrintDialog: TPrintDialog;
    FileSeed: TAction;
    SetMaze1: TMenuItem;
    SetMaze2: TMenuItem;
    FilePrintSetup: TAction;
    FileNew: TAction;
    N3: TMenuItem;
    PrinterSetup1: TMenuItem;
    GenerateNewMaze1: TMenuItem;
    Options: TAction;
    ZoomIn: TAction;
    ZoomOut: TAction;
    N2: TMenuItem;
    ZoomIn1: TMenuItem;
    ZoomOut1: TMenuItem;
    N15: TMenuItem;
    ZoomIn2: TMenuItem;
    ZoomOut2: TMenuItem;
    oggleTraditionalFlag1: TMenuItem;
    OptIsTraditional: TAction;
    OptShowLetters: TAction;
    OptShowMessage: TAction;
    OptShowSolution: TAction;
    OptSetMessage: TAction;
    ShowLetters1: TMenuItem;
    ShowMessage1: TMenuItem;
    ShowSolution1: TMenuItem;
    SetMessage1: TMenuItem;
    OptionsMainMenu: TMenuItem;
    raditionalMaze1: TMenuItem;
    ShowLetters2: TMenuItem;
    ShowMessage2: TMenuItem;
    ShowSolution2: TMenuItem;
    N4: TMenuItem;
    DisplayOptions1: TMenuItem;
    N5: TMenuItem;
    SetMessage2: TMenuItem;
    OptSetHilight: TAction;
    HighlightMessage1: TMenuItem;
    SetHilight2: TMenuItem;
    Recenter: TAction;
    Recenter1: TMenuItem;
    Recenter2: TMenuItem;
    procedure FilePrintExecute(Sender: TObject);
    procedure FileExitExecute(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure FileSeedExecute(Sender: TObject);
    procedure FilePrintSetupExecute(Sender: TObject);
    procedure FileNewExecute(Sender: TObject);
    procedure ZoomInExecute(Sender: TObject);
    procedure ZoomOutExecute(Sender: TObject);
    procedure OptIsTraditionalExecute(Sender: TObject);
    procedure OptShowLettersExecute(Sender: TObject);
    procedure OptShowMessageExecute(Sender: TObject);
    procedure OptShowSolutionExecute(Sender: TObject);
    procedure OptSetMessageExecute(Sender: TObject);
    procedure OptSetHilightExecute(Sender: TObject);
    procedure OptionsExecute(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure RecenterExecute(Sender: TObject);
  private
  	{ Keep track of the mouse }
    MouseDownFlag: Boolean;
    MouseLastX: Integer;
    MouseLastY: Integer;

	{ Display Flags }
    ISTRADMAZE:   Boolean;
    ISLETTERS:    Boolean;
    ISMESSAGE:    Boolean;
    ISHILIGHT:    Boolean;
    ISSOLSHOWING: Boolean;

    { Other Variables and Functions }
    MazeWalls:  AAWallExists;
	MazeSol:    TList;
    MazeSolStep:Array of Array of Integer;
	BaseName:   String;
    MazeMsg:    String;
    PaddedMsg:  String;
    SolLength:  Integer;
	SeedVal:    LongInt;
	ZoomFactor: Integer;
    XOffset, YOffset: Integer;
	DrawMutex: Boolean;
	HasInitOptsDlg: Boolean;
    function  CalculateMaxPrinterScale: Integer;
	function  DoPadMessage: String;
    function  FirstInPadMsg: Integer;
	procedure RecordSolution;
    procedure AutoCorrectSettings;
  	procedure Draw(MyCanvas: TCanvas; ScaleFactor: Integer);
	procedure InitMazeOnCanvas(MyCanvas: TCanvas; ScaleFactor: Integer);
    procedure DrawTraditionalMaze(MyCanvas: TCanvas; ScaleFactor: Integer);
    procedure DrawCorridorMaze(MyCanvas: TCanvas; ScaleFactor: Integer);
    procedure DrawSolution(MyCanvas: TCanvas; ScaleFactor: Integer);
	procedure DrawLetters(MyCanvas: TCanvas; ScaleFactor: Integer);
	procedure RefreshCaption;
  public
	{ Maze Display Parameters }
	W: Integer;  { Number of Boxes, Width and Height }
    H: Integer;
    WI: Integer; { Interior Height and Width; i.e.: each box }
    HI: Integer;
    PICLEFT: Integer;
    PICTOP: Integer;
    PICTAIL: Integer;
    PICPENWIDTH: Integer;
    AUTOCORRECT: Boolean;
    ALGORITHM: TMazeAlg;
  end;

var
  WordMazeForm: TWordMazeForm;

implementation

uses WordMazeOptions;

{$R *.dfm}

const
	MAXZOOM = 100;
    { Letter frequency computed from "Around the World in Eighty Days" }
    { With compliments to Jules Verne and the Public Domain, and to
      Project Gutenburg for the source text. }
    LetterFreq: Array[0..25] of Extended = (
		0.0839, { 'a' }
		0.0148, { 'b' }
		0.0260, { 'c' }
		0.0444, { 'd' }
		0.1244, { 'e' }
		0.0242, { 'f' }
		0.0236, { 'g' }
		0.0620, { 'h' }
		0.0659, { 'i' }
		0.0014, { 'j' }
		0.0063, { 'k' }
		0.0368, { 'l' }
		0.0238, { 'm' }
		0.0667, { 'n' }
		0.0751, { 'o' }
		0.0217, { 'p' }
		0.0010, { 'q' }
		0.0594, { 'r' }
		0.0654, { 's' }
		0.0923, { 't' }
		0.0288, { 'u' }
		0.0098, { 'v' }
		0.0225, { 'w' }
		0.0028, { 'x' }
		0.0165, { 'y' }
		0.0005  { 'z' }
	);

	LetterFreqSum: Array[0..25] of Extended = (
		0.0839, { 'a' }
		0.0987, { 'b' }
		0.1247, { 'c' }
		0.1691, { 'd' }
		0.2935, { 'e' }
		0.3177, { 'f' }
		0.3413, { 'g' }
		0.4032, { 'h' }
		0.4691, { 'i' }
		0.4705, { 'j' }
		0.4768, { 'k' }
		0.5136, { 'l' }
		0.5374, { 'm' }
		0.6041, { 'n' }
		0.6793, { 'o' }
		0.7009, { 'p' }
		0.7020, { 'q' }
		0.7614, { 'r' }
		0.8268, { 's' }
		0.9191, { 't' }
		0.9479, { 'u' }
		0.9577, { 'v' }
		0.9802, { 'w' }
		0.9830, { 'x' }
		0.9995, { 'y' }
		1.0000  { 'z' }
	);

    SpaceChance = 0.035;

function TWordMazeForm.CalculateMaxPrinterScale: Integer;
var
	MaxScaleX, MaxScaleY: Integer;
begin
	MaxScaleX := Printer.PageWidth  div (2*PICLEFT + W*WI);
	MaxScaleY := Printer.PageHeight div (2*PICTOP  + H*HI);
	if MaxScaleX < MaxScaleY then
    	Result := MaxScaleX
    else
    	Result := MaxScaleY;
end;

function TWordMazeForm.DoPadMessage: String;
var
	NPadLeft, NPadRight: Integer;
    PadLeft, PadRight: String;
    i, k: Integer;
    rnd: Extended;
    Letter: Char;
begin
	Result := '';
	if SolLength = Length(MazeMsg) then { return the message as-is }
    	Result := MazeMsg
    else if SolLength < Length(MazeMsg) then { chop the message to fit the space }
		Result := Copy(MazeMsg, 0, SolLength)
    else if SolLength > Length(MazeMsg) then
    begin
    	NPadLeft  := (SolLength - Length(MazeMsg)) div 2;
        NPadRight := SolLength - NPadLeft - Length(MazeMsg);
        PadLeft   := '';
        PadRight  := '';

		for i := 1 to NPadLeft - 1 do
    	begin
			rnd := Random;
        	letter := #0;
        	for k := Low(LetterFreqSum) to High(LetterFreqSum) do { [0..26] }
        	begin
				if rnd < LetterFreqSum[k] then
            	begin
                	letter := Chr(k + Ord('a'));
                	Break;
            	end;
        	end;
            if Random < SpaceChance then letter := ' ';
            PadLeft := PadLeft + letter;
        end;
        if NPadLeft > 0 then PadLeft := PadLeft + ' ';

		for i := 1 to NPadRight - 1 do
    	begin
			rnd := Random;
        	letter := #0;
        	for k := Low(LetterFreqSum) to High(LetterFreqSum) do { [0..26] }
        	begin
				if rnd < LetterFreqSum[k] then
            	begin
                	letter := Chr(k + Ord('a'));
                	Break;
            	end;
        	end;
            if Random < SpaceChance then letter := ' ';
            PadRight := PadRight + letter;
        end;
        if NPadRight > 0 then PadRight := ' ' + PadRight;

        Result := PadLeft + MazeMsg + PadRight;
    end;
end;

function TWordMazeForm.FirstInPadMsg: Integer;
begin
    if SolLength <= Length(MazeMsg) then
		Result := 1
    else
    	Result := ((SolLength - Length(MazeMsg)) div 2) + 1;
end;

procedure TWordMazeForm.RecordSolution;
var
	i, step : Integer;
    PCurPoint: PMazePointEx;
begin
	step := 1;
	for i := MazeSol.Count - 1 downto 0 do
    begin
    	PCurPoint := MazeSol.Items[i];
        MazeSolStep[PCurPoint^.Y, PCurPoint^.X] := step;
        step := step + 1;
    end;
end;

procedure TWordMazeForm.FormCreate(Sender: TObject);
var
	i, j: Integer;
begin
	BaseName := Caption;
	Randomize;
    SeedVal := RandSeed;
    Caption := BaseName + ' - Maze #' + IntToStr(SeedVal);

    ZoomFactor := 1;

	{ Set the maze parameters }
	W := 24;  { Number of Boxes, Width and Height }
    H := 18;
    WI := 36 ; { Interior Height and Width; i.e.: each box }
    HI := 36;
    PICPENWIDTH := 3;
    PICLEFT := 5 + WI div 2;
    PICTOP := 5 + HI div 2;
    PICTAIL := 3;
    ALGORITHM := mzHunt;
    AUTOCORRECT := True;

    ISTRADMAZE := True;
	ISLETTERS  := False;
    ISMESSAGE  := False;
    ISSOLSHOWING := False;

    HasInitOptsDlg := False;

    XOffset := 0;
    YOffset := 0;

    MazeMsg := '';

	SetLength(MazeWalls, H, W);
	for j := Low(MazeWalls) to High(MazeWalls) do
		for i := Low(MazeWalls[j]) to High(MazeWalls[j]) do
        	with MazeWalls[j,i] do
        	begin
        		RightWall  := False;
            	BottomWall := False;
        	end;

    SetLength(MazeSolStep, H, W);
    for j := Low(MazeSolStep) to High(MazeSolStep) do
    	for i := Low(MazeSolStep[j]) to High(MazeSolStep[j]) do
        	MazeSolStep[j,i] := 0;

	MouseDownFlag := False;
    DrawMutex := False;
end;

procedure TWordMazeForm.FilePrintExecute(Sender: TObject);
var
	TmpOffsetX, TmpOffsetY: Integer;
begin
	if PrintDialog.Execute then
    begin
    	{ShowMessage('(' + IntToStr(Printer.PageWidth) + ', ' +
        		IntToStr(Printer.PageHeight) + ')');}
        TmpOffsetX := XOffset;
        TmpOffsetY := YOffset;
    	XOffset := 0;
    	YOffset := 0;
        Printer.BeginDoc;
        Draw(Printer.Canvas, CalculateMaxPrinterScale);
        Printer.EndDoc;
		XOffset := TmpOffsetX;
        YOffset := TmpOffsetY;
    end;
end;

procedure TWordMazeForm.FilePrintSetupExecute(Sender: TObject);
begin
	PrintDialog.Execute;
    ShowMessage('Printer Scale: ' + inttostr(calculatemaxprinterscale));
end;

procedure TWordMazeForm.FileExitExecute(Sender: TObject);
begin
	Close;
end;

procedure TWordMazeForm.FormPaint(Sender: TObject);
begin
	Draw(Canvas, ZoomFactor);
end;

procedure TWordMazeForm.Draw(MyCanvas: TCanvas; ScaleFactor: Integer);
begin
	if DrawMutex then Exit;

    DrawMutex := True;
    try

	RandSeed := SeedVal;

    InitMazeOnCanvas(MyCanvas, ScaleFactor);

    MazeSol := TList.Create;
    try
    	GenerateMaze(MazeWalls, MazeSol, ALGORITHM);
        SolLength := MazeSol.Count;

        RecordSolution;
        if ISMESSAGE then
        	PaddedMsg := DoPadMessage;

    	if ISTRADMAZE then
			DrawTraditionalMaze(MyCanvas, ScaleFactor)
    	else
    		DrawCorridorMaze(MyCanvas, ScaleFactor);

        if ISSOLSHOWING then
        	DrawSolution(MyCanvas, ScaleFactor);

        if ISLETTERS then
        	DrawLetters(MyCanvas, ScaleFactor);


    finally
    	while MazeSol.Count > 0 do
        begin
        	Dispose(MazeSol.Items[0]);
            MazeSol.Delete(0);
        end;
    	MazeSol.Free;
        RefreshCaption;
    end;

    finally
    	DrawMutex := False;
    end;

end;

procedure TWordMazeForm.InitMazeOnCanvas(MyCanvas: TCanvas; ScaleFactor: Integer);
var
	i, j: Integer;
begin
	{ Initialize the maze to have walls everywhere }
	SetLength(MazeWalls, H, W);
    for j := 0 to (H - 1) do
    	for i := 0 to (W - 1) do
        	with MazeWalls[j,i] do
        	begin
        		{RightWall  := (Random > 0.5);}
            	{BottomWall := (Random > 0.5);}
                RightWall  := False;
                BottomWall := False;
        	end;
end;

procedure TWordMazeForm.DrawTraditionalMaze(MyCanvas: TCanvas; ScaleFactor: Integer);
var
	i, j : Integer;
    x, y : Integer;
    sc   : Integer;
begin
	sc := ScaleFactor;
    x := PICLEFT*sc + XOffset;
    y := PICTOP*sc + YOffset;
    with MyCanvas do
    begin
    	Pen.Width := sc*PICPENWIDTH;
        Pen.Color := clBlack;
    	Brush.Color := clWhite;
	    MoveTo(x, y + HI*sc);
        LineTo(x, y + H*HI*sc);
        LineTo(x + W*WI*sc, y + H*HI*sc);
        MoveTo(x + W*WI*sc, y + H*HI*sc - HI*sc);
        LineTo(x + W*WI*sc, y);
        LineTo(x, y);
    end;
	for j := 0 to (H - 1) do
    begin
		for i := 0 to (W - 1) do
       	begin
       		x := PICLEFT*sc + WI*sc*i + XOffset;
           	y := PICTOP*sc  + HI*sc*j + YOffset;
           	if not ((j = (H - 1)) or (MazeWalls[j,i].RightWall)) then
           	begin
           		MyCanvas.MoveTo(x, y + HI*sc);
               	MyCanvas.LineTo(x + WI*sc, y + HI*sc);
           	end;
           	if not ((i = (W - 1)) or (MazeWalls[j,i].BottomWall)) then
           	begin
           		MyCanvas.MoveTo(x + WI*sc, y);
               	MyCanvas.LineTo(x + WI*sc, y + HI*sc);
           	end;
       	end;
    end;
end;

procedure TWordMazeForm.DrawCorridorMaze(MyCanvas: TCanvas; ScaleFactor: Integer);
var
	i, j : Integer;
    x, y : Integer;
	r    : TRect;
    Tail : Integer;
    sc   : Integer;
    pts  : Array[1..5] of TPoint;
begin
	sc := ScaleFactor;
	MyCanvas.Pen.Width := sc*PICPENWIDTH;
	MyCanvas.Pen.Color := clBlack;
    MyCanvas.Brush.Color := clWhite;
    Tail := PICTAIL*sc;
	for j := 0 to (H - 1) do
    begin
		for i := 0 to (W - 1) do
        begin
        	x := PICLEFT*sc + WI*sc*i + XOffset;
            y := PICTOP*sc  + HI*sc*j + YOffset;
            pts[1].X := x;
            pts[1].Y := y;
            pts[2].X := x;
            pts[2].Y := y + HI*sc;
            pts[3].X := x + WI*sc;
            pts[3].Y := y + HI*sc;
            pts[4].X := x + WI*sc;
            pts[4].Y := y;
            pts[5] := Pts[1];
        	MyCanvas.PolyLine(pts);
        end;
    end;

	MyCanvas.Pen.Color   := clWhite;
    MyCanvas.Brush.Color := clWhite;
	for j := 0 to (H - 1) do
    begin
		for i := 0 to (W - 1) do
        begin
        	x := PICLEFT*sc + WI*sc*i + XOffset;
            y := PICTOP*sc  + HI*sc*j + YOffset;
			with MyCanvas do
            begin
            	if (j <> (H - 1)) and (MazeWalls[j,i].RightWall) then
                begin
               		r.Left   := x +    WI*sc div 3;
               		r.Right  := x + 2*(WI*sc div 3);
                	r.Top    := y + HI*sc - Pen.Width;
                	r.Bottom := y + HI*sc + Pen.Width;
                    Rectangle(r);
                    Pen.Color := clBlack;
                    MoveTo(x + WI*sc div 3, y + HI*sc - Tail);
					LineTo(x + WI*sc div 3, y + HI*sc + Tail);
                    MoveTo(x + 2*(WI*sc div 3), y + HI*sc - Tail);
					LineTo(x + 2*(WI*sc div 3), y + HI*sc + Tail);
                    Pen.Color := clWhite;
                end;
            	if (i <> (W - 1)) and (MazeWalls[j,i].BottomWall) then
                begin
               		r.Top    := y +    HI*sc div 3;
               		r.Bottom := y + 2*(HI*sc div 3);
                	r.Left   := x + WI*sc - Pen.Width;
                	r.Right  := x + WI*sc + Pen.Width;
                    Rectangle(r);
                    Pen.Color := clBlack;
                    MoveTo(x + WI*sc - Tail, y + HI*sc div 3);
					LineTo(x + WI*sc + Tail, y + HI*sc div 3);
                    MoveTo(x + WI*sc - Tail, y + 2*(HI*sc div 3));
					LineTo(x + WI*sc + Tail, y + 2*(HI*sc div 3));
                	Pen.Color := clWhite;
                end;
            end;
        end;
    end;

    with MyCanvas do
    begin
    	x := PICLEFT*sc - WI*sc + XOffset;
		y := PICTOP*sc + YOffset;
		r.Top    := y +    HI*sc div 3;
    	r.Bottom := y + 2*(HI*sc div 3);
    	r.Left   := x + WI*sc - Pen.Width;
    	r.Right  := x + WI*sc + Pen.Width;
    	Rectangle(r);
    	Pen.Color := clBlack;
    	MoveTo(x + WI*sc - Tail, y + HI*sc div 3);
		LineTo(x + WI*sc + Tail, y + HI*sc div 3);
    	MoveTo(x + WI*sc - Tail, y + 2*(HI*sc div 3));
		LineTo(x + WI*sc + Tail, y + 2*(HI*sc div 3));
    	Pen.Color := clWhite;

    	x := PICLEFT*sc + WI*(W - 1)*sc + XOffset;
		y := PICTOP*sc  + HI*(H - 1)*sc + YOffset;
		r.Top    := y +    HI*sc div 3;
    	r.Bottom := y + 2*(HI*sc div 3);
    	r.Left   := x + WI*sc - Pen.Width;
    	r.Right  := x + WI*sc + Pen.Width;
    	Rectangle(r);
    	Pen.Color := clBlack;
    	MoveTo(x + WI*sc - Tail, y + HI*sc div 3);
		LineTo(x + WI*sc + Tail, y + HI*sc div 3);
    	MoveTo(x + WI*sc - Tail, y + 2*(HI*sc div 3));
		LineTo(x + WI*sc + Tail, y + 2*(HI*sc div 3));
    	Pen.Color := clWhite;
	end;
end;

procedure TWordMazeForm.DrawSolution(MyCanvas: TCanvas; ScaleFactor: Integer);
var
	i, sc: Integer;
    x, y:  Integer;
    CurPoint: PMazePointEx;
begin
	sc := ScaleFactor;
	if MazeSol.Count <= 0 then Exit;

	MyCanvas.Pen.Width := sc*PICPENWIDTH;
	MyCanvas.Pen.Color := clGray;

	{ Enter the maze }
    CurPoint := MazeSol.Items[MazeSol.Count - 1];
    x := PICLEFT*sc + WI*CurPoint^.X*sc - WI*sc + XOffset;
    y := PICTOP*sc  + HI*CurPoint^.Y*sc + YOffset;
    MyCanvas.MoveTo(x + WI*sc div 2, y + HI*sc div 2);

	{ Draw the solution of the maze }
	for i := MazeSol.Count - 1 downto 0 do
    begin
    	CurPoint  := MazeSol.Items[i];
    	x := PICLEFT*sc + WI*CurPoint^.X*sc + XOffset;
		y := PICTOP*sc  + HI*CurPoint^.Y*sc + YOffset;
        MyCanvas.LineTo(x + WI*sc div 2, y + HI*sc div 2);
    end;

    { Exit the maze }
    CurPoint := MazeSol.Items[0];
    x := PICLEFT*sc + WI*CurPoint^.X*sc + WI*sc + XOffset;
    y := PICTOP*sc  + HI*CurPoint^.Y*sc + YOffset;
    MyCanvas.LineTo(x + WI*sc div 2, y + HI*sc div 2);

end;

procedure TWordMazeForm.DrawLetters(MyCanvas: TCanvas; ScaleFactor: Integer);
var
	n_letter: Byte;
    letter: Char;
	rnd: Extended;
	b, IsOnPath: Boolean;
	i, j, k, x, y, tw, th, sc: Integer;
    StartMsg, EndMsg: Integer;
    PCurPoint: PMazePointEx;
begin
	for j := 0 to H - 1 do
    begin
		for i := 0 to W - 1 do
        begin
			{ Choose the letter according to its frequency in English }
			rnd := Random;
            letter := #0;
            for k := Low(LetterFreqSum) to High(LetterFreqSum) do { [0..26] }
            begin
				if rnd < LetterFreqSum[k] then
                begin
                	n_letter := k;
                    letter := Chr(n_letter + Ord('a'));
                    Break;
                end;
            end;
            if Random < SpaceChance then letter := ' ';
            if letter = #0 then
            begin
            	ShowMessage('Internal Error #1');
                Exit;
            end;

			{ Check if the letter is on the solution path }
            IsOnPath := False;
            b := False;
            for k := 0 to MazeSol.Count - 1 do
            begin
            	PCurPoint := MazeSol.Items[k];
                if (PCurPoint^.X = i) and (PCurPoint^.Y = j) then
                	b := True;
            end;
            if b then
            begin
            	{ If there is a message, don't put a random letter in it! }
            	if ISMESSAGE and (Length(Trim(MazeMsg)) > 0) then
            		letter := (Copy(PaddedMsg, MazeSolStep[j, i], 1))[1];
            	IsOnPath := True;
            end;

            StartMsg := FirstInPadMsg;
            EndMsg   := StartMsg + Length(MazeMsg);

            sc := ScaleFactor;
    		x := PICLEFT*sc + WI*i*sc + XOffset;
    		y := PICTOP*sc  + HI*j*sc + YOffset;
            MyCanvas.Brush.Style := bsClear;
            MyCanvas.Font.Name := 'Courier New';
            if not (ISMESSAGE and ISHILIGHT and ISSOLSHOWING and IsOnPath and
        			(Length(MazeMsg) > 0) and
                    (MazeSolStep[j, i] in [StartMsg..EndMsg])) then
	            MyCanvas.Font.Height := -HI*sc div 3
            else
	            MyCanvas.Font.Height := -2*HI*sc div 3;
    		MyCanvas.Font.Style := [fsBold];
			MyCanvas.Font.Color := clBlack;
            tw := MyCanvas.TextWidth(letter);
            th := MyCanvas.TextHeight(letter);
            x := x + (WI*sc - tw) div 2;
            y := y + (HI*sc - th) div 2;
            if ISMESSAGE and ISHILIGHT and IsOnPath and (Length(MazeMsg) > 0) then
            begin
				if (MazeSolStep[j, i] in [StartMsg..EndMsg]) then
            		MyCanvas.Font.Color := clRed
                else
                	MyCanvas.Font.Color := clBlack;
                MyCanvas.TextOut(x, y, letter);
                MyCanvas.Font.Color := clBlack;
            end
            else
            	MyCanvas.TextOut(x, y, letter);


        end;
    end;
end;


procedure TWordMazeForm.FileSeedExecute(Sender: TObject);
var
	RetStr: String;
begin
	RetStr := IntToStr(SeedVal);
	if InputQuery('Select Maze', 'Enter New Maze Number', RetStr) then
    begin
    	try
        	if LeftStr(Trim(LowerCase(RetStr)), 4) <> 'rand' then
				SeedVal := StrToInt(RetStr)
            else
            	SeedVal := RandSeed;
        except
        	on EConvertError do begin
            	ShowMessage('Error! Invalid integer entered!  Please try again.');
                Exit;
            end;
        end;
        RandSeed := SeedVal;
        Refresh;
        RefreshCaption;
    end;
end;

procedure TWordMazeForm.FileNewExecute(Sender: TObject);
begin
	SeedVal := RandSeed;
    Refresh;
    RefreshCaption;
end;

procedure TWordMazeForm.RefreshCaption;
begin
    Caption := BaseName + ' - Maze #' + IntToStr(SeedVal);
    if not ISMESSAGE then
    	Caption := Caption + ' (Solution Length: ' + IntToStr(SolLength) + ')'
    else
    	Caption := Caption + ' (Message: ' + PaddedMsg + ')';
    if ZoomFactor <> 1 then
    	Caption := Caption + ' (Zoom: ' + IntToStr(ZoomFactor) + 'x)';
end;

procedure TWordMazeForm.ZoomInExecute(Sender: TObject);
begin
	if ZoomFactor < MAXZOOM then
    begin
    	ZoomFactor := ZoomFactor + 1;
        Refresh;
        RefreshCaption;
    end;
end;

procedure TWordMazeForm.ZoomOutExecute(Sender: TObject);
begin
	if ZoomFactor > 1 then
    begin
    	ZoomFactor := ZoomFactor - 1;
        Refresh;
        RefreshCaption;
    end;
end;

procedure TWordMazeForm.OptIsTraditionalExecute(Sender: TObject);
begin
	ISTRADMAZE := not ISTRADMAZE;
	if AUTOCORRECT then AutoCorrectSettings;
    Refresh;
end;

procedure TWordMazeForm.OptShowLettersExecute(Sender: TObject);
begin
	ISLETTERS := not ISLETTERS;
    OptShowMessage.Enabled := ISLETTERS;
    OptSetMessage.Enabled := ISLETTERS and ISMESSAGE;
    OptSetHilight.Enabled := ISLETTERS and ISMESSAGE;
    Refresh;
end;

procedure TWordMazeForm.OptShowMessageExecute(Sender: TObject);
begin
	if not ISMESSAGE then
    begin
    	if Trim(MazeMsg) = '' then OptSetMessageExecute(Sender);
    	if Trim(MazeMsg) = '' then Exit;
    end;

	ISMESSAGE := not ISMESSAGE;
    OptSetMessage.Enabled := ISMESSAGE;
    OptSetHilight.Enabled := ISMESSAGE;
    Refresh;
end;

procedure TWordMazeForm.OptShowSolutionExecute(Sender: TObject);
begin
	ISSOLSHOWING := not ISSOLSHOWING;
    Refresh;
end;

procedure TWordMazeForm.OptSetMessageExecute(Sender: TObject);
var
	RetStr: String;
begin
	RetStr := MazeMsg;
	if InputQuery('Message', 'Enter Solution Message', RetStr) then
    begin
    	MazeMsg := RetStr;
    	Refresh;
    end;
end;

procedure TWordMazeForm.OptSetHilightExecute(Sender: TObject);
begin
	ISHILIGHT := not ISHILIGHT;
    Refresh;
end;

procedure TWordMazeForm.OptionsExecute(Sender: TObject);
var
	i, j: Integer;
	OldW, OldH: Integer;
begin
	if not HasInitOptsDlg then
    begin
    	HasInitOptsDlg := True;
        WordMazeOpts.MazeAlgorithm.AddItem('Hunt Algorithm', nil);
        WordMazeOpts.MazeAlgorithm.AddItem('Recursive Backtracking (DFS)', nil);
        WordMazeOpts.MazeAlgorithm.AddItem('Prim''s Algorithm (BFS)', nil);
    end;

	OldW := W;
    OldH := H;
	WordMazeOpts.WSet.Position := W;
	WordMazeOpts.HSet.Position := H;
	WordMazeOpts.WISet.Position := WI;
	WordMazeOpts.HISet.Position := HI;
	WordMazeOpts.PicLeftSet.Position := PICLEFT;
	WordMazeOpts.PicTopSet.Position := PICTOP;
	WordMazeOpts.PicTailSet.Position := PICTAIL;
	WordMazeOpts.PicPenWidthSet.Position := PICPENWIDTH;
    WordMazeOpts.AutoCorrectSet.Checked := AUTOCORRECT;
    case ALGORITHM of
    	mzHunt: WordMazeOpts.MazeAlgorithm.ItemIndex := 0;
        mzBacktrack: WordMazeOpts.MazeAlgorithm.ItemIndex := 1;
        mzPrim: WordMazeOpts.MazeAlgorithm.ItemIndex := 2;
    else
    	WordMazeOpts.MazeAlgorithm.ItemIndex := -1;
    end;
	if IsPositiveResult(WordMazeOpts.ShowModal) then
	begin
    	case WordMazeOpts.MazeAlgorithm.ItemIndex of
        	0: ALGORITHM := mzHunt;
            1: ALGORITHM := mzBacktrack;
            2: ALGORITHM := mzPrim;
        else
        	ALGORITHM := mzHunt;
        end;
		W := WordMazeOpts.WSet.Position;
		H := WordMazeOpts.HSet.Position;
		WI := WordMazeOpts.WISet.Position;
		HI := WordMazeOpts.HISet.Position;
		PICLEFT := WordMazeOpts.PicLeftSet.Position;
		PICTOP := WordMazeOpts.PicTopSet.Position;
		PICTAIL := WordMazeOpts.PicTailSet.Position;
		PICPENWIDTH := WordMazeOpts.PicPenWidthSet.Position;
        AUTOCORRECT := WordMazeOpts.AutoCorrectSet.Checked;
        if AUTOCORRECT then AutoCorrectSettings;
        if (W <> OldW) or (H <> OldH) then
        begin
        	SetLength(MazeWalls, H, W);
			for j := Low(MazeWalls) to High(MazeWalls) do
				for i := Low(MazeWalls[j]) to High(MazeWalls[j]) do
        			with MazeWalls[j,i] do
        			begin
        				RightWall  := False;
            			BottomWall := False;
        			end;

    		SetLength(MazeSolStep, H, W);
			for j := Low(MazeSolStep) to High(MazeSolStep) do
				for i := Low(MazeSolStep[j]) to High(MazeSolStep[j]) do
        			MazeSolStep[j,i] := 0;

        	XOffset := 0;
    		YOffset := 0;

        end;
    	Refresh;
	end;
end;

procedure TWordMazeForm.AutoCorrectSettings;
begin
	if ISTRADMAZE then
    begin
    	if (PICPENWIDTH > (WI div 2)) then
        	PICPENWIDTH := WI div 2;
        if (PICPENWIDTH > (HI div 2)) then
        	PICPENWIDTH := HI div 2;
        if (PICTAIL > (WI div 6)) then
        	PICTAIL := WI div 6;
        if (PICTAIL > (HI div 6)) then
        	PICTAIL := HI div 6;
    end
    else
    begin
    	if (PICPENWIDTH > (WI div 6)) then
        	PICPENWIDTH := WI div 6;
        if (PICPENWIDTH > (HI div 6)) then
        	PICPENWIDTH := HI div 6;
        if (PICTAIL > (WI div 6)) then
        	PICTAIL := WI div 6;
        if (PICTAIL > (HI div 6)) then
        	PICTAIL := HI div 6;
	end;
end;

procedure TWordMazeForm.FormMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
	if Button = mbLeft then
	begin
    	MouseDownFlag := True;
        MouseLastX := X;
        MouseLastY := Y;
    end;
end;

procedure TWordMazeForm.FormMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
	if MouseDownFlag then
    begin
		MouseDownFlag := False;
    end;
end;

procedure TWordMazeForm.FormMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
var
	OldXOffset, OldYOffset: Integer;
begin
	if MouseDownFlag then
    begin
    	OldXOffset := XOffset;
        OldYOffset := YOffset;
    	XOffset := XOffset + (X - MouseLastX);
    	YOffset := YOffset + (Y - MouseLastY);
        if (XOffset <> OldXOffset) or (YOffset <> OldYOffset) then
        	Refresh;
        MouseLastX := X;
        MouseLastY := Y;
    end;
end;

procedure TWordMazeForm.RecenterExecute(Sender: TObject);
begin
	XOffset := 0;
    YOffset := 0;
    Refresh;
end;

end.
